from unittest import TestCase

from uhlive.stream.conversation import Conversation, Ok, ProtocolError

from .conversation_events import join_successful


class TestConnection(TestCase):
    def test_join(self):
        client = Conversation("customerid", "myconv", "john_test")
        frame = client.join(model="en", country="us")
        self.assertEqual(
            frame,
            r'["1","1","conversation:customerid@myconv","phx_join",{"readonly":false,"speaker":"john_test","model":"en","country":"us","interim_results":true,"rescoring":true,"origin":0,"audio_codec":"linear"}]',
        )
        with self.assertRaises(ProtocolError):
            client.send_audio_chunk(bytes(60))

        with self.assertRaises(ProtocolError):
            client.leave()

    def test_joined(self):
        client = Conversation("customerid", "myconv", "john_test")
        client.join()
        event = client.receive(join_successful)
        self.assertIsInstance(event, Ok)

        # Can't join twice
        with self.assertRaises(ProtocolError):
            client.join()

        # Can stream
        client.send_audio_chunk(bytes(60))
        # Can leave
        client.leave()

    def test_leave(self):
        client = Conversation("customerid", "myconv", "john_test")
        client.join()
        client.receive(join_successful)

        frame = client.leave()
        self.assertEqual(
            frame, r'["1","2","conversation:customerid@myconv","phx_leave",{}]'
        )

    def receive_wrong_topic(self):
        client = Conversation("customerid", "unrelated_topic", "john_test")
        with self.assertRaises(AssertionError):
            client.receive(join_successful)



from datetime import datetime
from unittest import TestCase

from uhlive.stream.recognition import events

from .recog_events import (
    grammar_defined,
    method_failed,
    params_set,
    recognition_complete,
    recognition_complete_nbests,
    session_opened,
)


class TestEventDeserialization(TestCase):
    def test_session_opened(self):
        event = events.deserialize(session_opened)
        self.assertIsInstance(event, events.Opened)
        self.assertEqual(event.request_id, 0)
        self.assertEqual(event.channel_id, "testuie46e4ui6")
        self.assertEqual(event.headers, {})
        self.assertIsNone(event.completion_cause)
        self.assertIsNone(event.completion_reason)
        self.assertIsNone(event.body)

    def test_grammar_defined(self):
        event = events.deserialize(grammar_defined)
        self.assertIsInstance(event, events.GrammarDefined)
        self.assertEqual(event.request_id, 1)
        self.assertEqual(event.channel_id, "testuie46e4ui6")
        self.assertEqual(event.headers, {})
        self.assertIsNone(event.completion_cause)
        self.assertIsNone(event.completion_reason)
        self.assertIsNone(event.body)

    def test_params_set(self):
        event = events.deserialize(params_set)
        self.assertIsInstance(event, events.ParamsSet)
        self.assertEqual(event.request_id, 2)
        self.assertEqual(event.channel_id, "testuie46e4ui6")
        self.assertEqual(event.headers, {})
        self.assertIsNone(event.completion_cause)
        self.assertIsNone(event.completion_reason)
        self.assertIsNone(event.body)

    def test_recognition_complete(self):
        event = events.deserialize(recognition_complete)
        self.assertIsInstance(event, events.RecognitionComplete)
        self.assertEqual(event.request_id, 3)
        self.assertEqual(event.channel_id, "testuie46e4ui6")
        self.assertEqual(event.headers, {})
        self.assertEqual(event.completion_cause, events.CompletionCause.Success)
        self.assertIsNone(event.completion_reason)
        result = event.body
        self.assertIsInstance(result, events.RecogResult)
        self.assertIsInstance(result.asr, events.Transcript)
        self.assertIsInstance(result.nlu, events.Interpretation)
        self.assertEqual(result.grammar_uri, "session:immat")
        self.assertEqual(result.nlu.value, "bc305fz")
        self.assertEqual(
            result.asr.transcript, "attendez alors voilà baissé trois cent cinq f z"
        )
        self.assertEqual(result.asr.start, datetime(2021, 8, 20, 10, 5, 34, 909000))

    def test_recognition_complete_nbests(self):
        event = events.deserialize(recognition_complete_nbests)
        result = event.body
        self.assertIsInstance(result, events.RecogResult)
        self.assertIsInstance(result.asr, events.Transcript)
        self.assertIsInstance(result.nlu, events.Interpretation)
        self.assertEqual(result.nlu.value, "le137137866cn")
        self.assertEqual(len(result.alternatives), 2)
        best2 = result.alternatives[0]
        self.assertIsInstance(best2, events.RecogResult)
        self.assertIsInstance(best2.asr, events.Transcript)
        self.assertIsInstance(best2.nlu, events.Interpretation)
        self.assertEqual(best2.nlu.value, "le137137866cm")

    def test_method_failed(self):
        event = events.deserialize(method_failed)
        self.assertIsInstance(event, events.MethodFailed)
        self.assertEqual(event.request_id, 2)
        self.assertEqual(event.channel_id, "testuie46e4ui6")
        self.assertEqual(event.headers, {})
        self.assertEqual(event.completion_cause, events.CompletionCause.GramLoadFailure)
        self.assertEqual(event.completion_reason, "unknown grammar 'toto'")

    def test_event_str(self):
        # We just want to be sure we don't raise exceptions
        event = events.deserialize(session_opened)
        print(event)
        event = events.deserialize(grammar_defined)
        print(event)
        event = events.deserialize(params_set)
        print(event)
        event = events.deserialize(recognition_complete)
        print(event)
        event = events.deserialize(method_failed)
        print(event)



"""
The Stream Conversation SDK API for human to human interactions.

This API is used to consume a real-time audio stream and get enriched transcription events.

The protocol is messages based and uses websockets as transport. You are free to use whatever websocket client library you like to communicate
with the API, and use our SDK to encode/decode the messages.

## Quickstart

First retrieve a one time access token with the [Auth API](auth.md).

Then use that token to build an authenticated URL, open a websocket connection to it with the websocket client library
of your choice and instanciate a [`Conversation`][uhlive.stream.conversation.Conversation] to join a conversation, generate
audio stream messages and decode transcription and enrichment events.

As the API is asynchronous, streaming the audio and reading the returned events should be done in two different threads/tasks.

```python
from uhlive.stream.Conversation import *

stream_h2h_url = build_conversation_url(token)

# The subcripttion identifier was given to you with your other credentials
# the conversation id can be any string you like. If a conversation by that name already exists in your subscription identifier domain
# it will join it as a new speaker, otherwise it will create it and join the speaker in.
# The speaker id helps you identify who is speaking.
conversation = Conversation("subscription_identifier", "a_conversation_id", "a_speaker_id")
```

Now you can connect and interact with the API:

Synchronous example:

```python
import websocket as ws

socket = ws.create_connection(stream_h2h_url, timeout=10)
socket.send(
    conversation.join(
        model="fr",
        interim_results=False,
        rescoring=True,
        origin=int(time.time() * 1000),
        country="fr",
    )
)
# check we didn't get an error on join
reply = conversation.receive(socket.recv())
assert isinstance(reply, Ok)

```

Asynchronous example:

```python
from aiohttp import ClientSession

async def main(uhlive_client, uhlive_secret):
    async with ClientSession() as session:
        async with session.ws_connect(stream_h2h_url) as socket:
            await socket.send_str(
                conversation.join(
                    model="fr",
                    interim_results=False,
                    rescoring=True,
                    origin=int(time.time() * 1000),
                    country="fr",
                )
            )
            # check we didn't get an error on join
            msg = await socket.receive()
            reply = conversation.receive(msg.data)
            assert isinstance(reply, Ok)
```

As you can see, the I/O is cleanly decoupled from the protocol handling: the `Conversation` object is only used
to create the messages to send to the API and to decode the received messages as `Event` objects.

See the [complete examples in the source distribution](https://github.com/uhlive/python-sdk/tree/main/examples/conversation).
"""

import os
from urllib.parse import urljoin

from .client import Conversation, ProtocolError
from .events import (
    EntityFound,
    EntityReference,
    Event,
    Ok,
    RelationFound,
    SegmentDecoded,
    SpeakerJoined,
    SpeakerLeft,
    SpeechDecoded,
    Tag,
    TagsFound,
    Unknown,
    Word,
    WordsDecoded,
)

SERVER = os.getenv("UHLIVE_API_URL", "wss://api.uh.live")


def build_conversation_url(token: str) -> str:
    """
    Make an authenticated URL to connect to the Conversation Service.
    """
    return urljoin(SERVER, "socket/websocket") + f"?jwt={token}&vsn=2.0.0"


__all__ = [
    "build_conversation_url",
    "Conversation",
    "ProtocolError",
    "SpeakerJoined",
    "Word",
    "EntityFound",
    "Event",
    "Ok",
    "EntityReference",
    "RelationFound",
    "SegmentDecoded",
    "SpeakerLeft",
    "SpeechDecoded",
    "Unknown",
    "WordsDecoded",
    "Tag",
    "TagsFound",
]



"""Event definitions."""

import re
from typing import Any, List

from .error import UhliveError
from .human_datetime import human_datetime

ENTITY_NAME = re.compile(r"entity_([\w_]+)_found")
RELATION_NAME = re.compile(r"relation_([\w_]+)_found")


class Word(dict):
    """Timestamped word."""

    @property
    def start(self) -> int:
        """Start time as Unix timestamp in millisecond, according to audio timeline."""
        return self["start"]

    @property
    def end(self) -> int:
        """End time as Unix timestamp in millisecond, according to audio timeline."""
        return self["end"]

    @property
    def length(self) -> int:
        """Word length in millisecond, according to audio timeline."""
        return self["length"]

    @property
    def word(self) -> str:
        """Transcript token string for this word."""
        return self["word"]

    @property
    def confidence(self) -> float:
        """ASR confidence for this word."""
        return self["confidence"]


class Event(object):
    """The base class of all events."""

    def __init__(self, join_ref, ref, topic, event, payload) -> None:
        self._join_ref = join_ref
        self._ref = ref
        self._topic = topic
        self._payload = payload

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(payload={self._payload})"

    @property
    def topic(self) -> str:
        """The conversation identifier"""
        return self._topic

    @property
    def join_ref(self) -> str:
        return self._join_ref

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def speaker(self) -> str:
        """The speaker whose speech triggered this event.

        All events are relative to a speaker."""
        return self._payload["speaker"]

    @staticmethod
    def from_message(message):
        """Private method to instantiate the right type of event from the raw websocket message."""
        [join_ref, ref, topic, event, payload] = message
        if event == "phx_reply" and payload["status"] == "error":
            raise UhliveError(payload["response"]["reason"])
        if event in EVENT_MAP:
            return EVENT_MAP[event](*message)
        if event.startswith("entity_"):
            return EntityFound(*message)
        if event.startswith("relation_"):
            return RelationFound(*message)
        if event == "phx_error":
            raise UhliveError("Server error, channel crashed!")
        return Unknown(*message)


class Ok(Event):
    """API asynchronous command aknowledgements."""

    pass


class Unknown(Event):
    """The server emitted an event unkown to this SDK. Time to upgrade!"""

    def __init__(self, join_ref, ref, topic, event, payload):
        self._name = event
        super().__init__(join_ref, ref, topic, event, payload)

    def __repr__(self) -> str:
        return f"Unknown [{self._name}](payload={self._payload})"


class TimeScopedEvent(Event):
    """Base class for events that are anchored to the audio time line."""

    @property
    def start(self) -> int:
        """Start time as Unix timestamp in millisecond, according to audio timeline."""
        return self._payload["start"]

    @property
    def end(self) -> int:
        """End time as Unix timestamp in millisecond, according to audio timeline."""
        return self._payload["end"]

    @property
    def length(self) -> int:
        """Event length in millisecond, according to audio timeline."""
        return self._payload["length"]


class SpeechDecoded(TimeScopedEvent):
    """The base class of all transcription events."""

    @property
    def transcript(self) -> str:
        """Get the transcript of the whole segment as a string"""
        return self._payload["transcript"]

    @property
    def lang(self) -> str:
        """Natural Language of the speech.

        As ISO 639-1 code.
        """
        return self._payload["lang"]

    @property
    def country(self) -> str:
        """Country location of speaker.

        As ISO 3166-1 code.
        """
        return self._payload["country"]

    @property
    def utterance_id(self) -> str:
        """The Utterance id identifies the speech utterance this event transcribes."""
        return self._payload["utterance_id"]

    @property
    def words(self) -> List[Word]:
        """Get the transcript of the whole segment as a list of timestamped [words][uhlive.stream.conversation.Word]."""
        return [Word(w) for w in self._payload["words"]]

    def __str__(self) -> str:
        return f"[{self.speaker} — {human_datetime(self.start)}] {self.transcript}"

    @property
    def confidence(self) -> float:
        """The ASR confidence for this segment."""
        return self._payload["confidence"]


class WordsDecoded(SpeechDecoded):
    """Interim segment transcript event."""

    pass


class SegmentDecoded(SpeechDecoded):
    """Final segment transcript event."""

    pass


class SegmentNormalized(SpeechDecoded):
    """Normalized final segment event."""

    def __str__(self) -> str:
        return f"[{self.speaker} — Formatted] {self.transcript}"


class SpeakerJoined(Event):
    """A new speaker joined the conversation (after us)."""

    @property
    def timestamp(self) -> int:
        """The UNIX time when the speaker joined the conversation."""
        return self._payload["timestamp"]

    @property
    def interim_results(self) -> bool:
        """Are interim results activated for this speaker?"""
        return self._payload["interim_results"]

    @property
    def rescoring(self) -> bool:
        """Is rescoring enabled for this speaker?"""
        return self._payload["rescoring"]

    @property
    def lang(self) -> str:
        """Natural Language of the speech.

        As ISO 639-1 code.
        """
        return self._payload["lang"]

    @property
    def country(self) -> str:
        """Country location of speaker.

        As ISO 3166-1 code.
        """
        return self._payload["country"]


class SpeakerLeft(Event):
    """Event emitted by the associated speaker when they left the conversation."""

    @property
    def timestamp(self) -> int:
        """UNIX time when the speaker left the conversation."""
        return self._payload["timestamp"]


class EntityFound(TimeScopedEvent):
    """The class for all entity annotation events."""

    def __init__(self, join_ref, ref, topic, event, payload):
        self._name = ENTITY_NAME.match(event).group(
            1
        )  # let it raise if it doesn't match
        super().__init__(join_ref, ref, topic, event, payload)

    @property
    def entity_name(self) -> str:
        """The name of the named entity found."""
        return self._name

    @property
    def lang(self) -> str:
        """Natural Language of the interpretation.

        As ISO 639-1 code.
        """
        return self._payload["lang"]

    @property
    def country(self) -> str:
        """Country location of speaker.

        As ISO 3166-1 code.
        """
        return self._payload["country"]

    @property
    def canonical(self) -> str:
        """The well formatted form of the entity in the language (string)."""
        return self._payload["annotation"].get("canonical")

    @property
    def original(self) -> str:
        """The transcript excerpt that was interpreted, as string."""
        return self._payload["annotation"]["original"]

    @property
    def value(self) -> Any:
        """The interpreted value in machine understandable form.

        The exact type depends on the entity.
        """
        return self._payload["annotation"].get("value")

    @property
    def confidence(self) -> float:
        """The confidence of the interpretation."""
        return self._payload["confidence"]

    def __repr__(self) -> str:
        return " ".join(
            (
                " - ",
                f"{self.__class__.__name__} in {self.speaker}:  <{self._name}> {self.canonical or self.original}",
                f"({self.value})" if self.value != self.canonical else "",
                f"[confidence: {self.confidence:.2f}]",
            )
        )


class Tag:
    """A tag represents a behavioral feature found in the conversation."""

    uuid: str
    """The unique id of the Tag."""
    label: str
    """The human readable name of the Tag."""

    def __init__(self, uuid: str, label: str) -> None:
        self.uuid = uuid
        self.label = label

    def __repr__(self) -> str:
        return f"Tag({self.label})"


class TagsFound(TimeScopedEvent):
    """One or more tags were found on this time range."""

    @property
    def lang(self) -> str:
        """Natural Language of the interpretation.

        As ISO 639-1 code.
        """
        return self._payload["lang"]

    @property
    def country(self) -> str:
        """Country location of speaker.

        As ISO 3166-1 code.
        """
        return self._payload["country"]

    @property
    def confidence(self) -> float:
        """Tagger confidence."""
        return self._payload["confidence"]

    @property
    def tags(self) -> List[Tag]:
        """The [tags][uhlive.stream.conversation.Tag] that were found on this time range"""
        return [Tag(t["uuid"], t["label"]) for t in self._payload["annotation"]["tags"]]

    def __repr__(self):
        return f"{self.__class__.__name__} in {self.speaker}:  {self.tags}"


class EntityReference:
    """Reference to a unique previously found Entity in the conversation."""

    kind: str
    """The name of the `Entity` referenced."""
    speaker: str
    """The speaker identifier."""
    start: int
    """The UNIX start time of the referenced `Entity`."""

    def __init__(self, entity_name: str, speaker: str, start: int) -> None:
        self.kind = entity_name
        self.speaker = speaker
        self.start = start

    def __repr__(self) -> str:
        return f"Ref({self.kind} by {self.speaker} at {human_datetime(self.start)})"


class RelationFound(TimeScopedEvent):
    """The class for all Relation events.

    Relations express a semantic relationship between two or more entities.
    """

    def __init__(self, join_ref, ref, topic, event, payload):
        self._name = RELATION_NAME.match(event).group(
            1
        )  # let it raise if it doesn't match
        super().__init__(join_ref, ref, topic, event, payload)

    @property
    def relation_name(self) -> str:
        """The type of the relation."""
        return self._name

    @property
    def lang(self) -> str:
        """Natural Language of the interpretation.

        As ISO 639-1 code.
        """
        return self._payload["lang"]

    @property
    def confidence(self) -> float:
        """The confidence on the discovered relationship."""
        return self._payload["confidence"]

    @property
    def members(self) -> List[EntityReference]:
        """[References to the Entities][uhlive.stream.conversation.EntityReference] involved in this relationship."""
        m = []
        speaker = self.speaker
        print(self._payload)
        for ref in self._payload["members"]:
            kind = (
                ENTITY_NAME.match(ref["entity"]).group(1) if ref["entity"] else None  # type: ignore
            )
            if kind is not None:
                m.append(EntityReference(kind, speaker, ref["start"]))
        return m

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <{self._name}> for {self.members} [confidence: {self.confidence:.2f}]"


EVENT_MAP = {
    "words_decoded": WordsDecoded,
    "segment_decoded": SegmentDecoded,
    "segment_normalized": SegmentNormalized,
    "tags_found": TagsFound,
    "speaker_joined": SpeakerJoined,
    "speaker_left": SpeakerLeft,
    "phx_reply": Ok,
    "phx_close": Ok,
}



"""
The stream recognition API SDK for voice bots.

Stream for voicebots, or Stream Human to Bots, or Stream H2B is a set of API enabling clients to build
interaction between a human end-user and a bot, for example to create Interactive Voice Response (IVR)
on the phone, or a voicebot within an app.

For an overview of the concepts, protocols and workflow, see the
[higher level documenation](https://docs.allo-media.net/stream-h2b/#real-time-stream-api-for-voicebots) and
more specifically the [Websocket H2B protocol reference](https://docs.allo-media.net/stream-h2b/protocols/websocket/#websocket-for-voicebots).

The protocol is messages based and uses websockets as transport. You are free to use whatever websocket client library you like to communicate
with the API, and use our SDK to encode/decode the messages.

## Quickstart

First retrieve a one time access token with the [Auth API](auth.md).

Then use that token to build an authenticated URL, open a websocket connection to it with the websocket client library
of your choice and instanciate a [`Recognizer`][uhlive.stream.recognition.Recognizer] to make request, generate
audio stream messages and decode responses.

As the API is asynchronous, streaming the audio and reading the returned events should be done in two different threads/tasks.

```python
from uhlive.stream.recognition import *

stream_h2b_url, stream_h2b_headers = build_connection_request(token)
recognizer = Recognizer()
```

Now you can connect and interact with the API:

Synchronous example:

```python
import websocket as ws

socket = ws.create_connection(stream_h2b_url, header=stream_h2b_headers)
socket.send(recognizer.open())
# Check if successful
reply = recognizer.receive(socket.recv())
assert isinstance(reply, Opened), f"Expected Opened, got {reply}"
# start streaming the user's voice in another thread
streamer_thread_handle = stream_mic(socket, recognizer)
```

Asynchronous example:

```python
from aiohttp import ClientSession

    async with ClientSession() as session:
        async with session.ws_connect(stream_h2b_url, header=stream_h2b_headers) as socket:

            # Open a session
            # Commands are sent as text frames
            await socket.send_str(recognizer.open())
            # Check if successful
            msg = await socket.receive()
            reply = recognizer.receive(msg.data)
            assert isinstance(reply, Opened), f"Expected Opened, got {reply}"
            # start streaming the user's voice in another task
            streamer_task_handle = asyncio.create_task(stream(socket, recognizer))
```

As you can see, the I/O is cleanly decoupled from the protocol handling: the `Recognizer` object is only used
to create the messages to send to the API and to decode the received messages as `Event` objects.

See the [complete examples in the source distribution](https://github.com/uhlive/python-sdk/tree/main/examples/recognition).

"""

import os
from typing import Tuple
from urllib.parse import urljoin

from .client import ProtocolError, Recognizer
from .events import (
    Closed,
    CompletionCause,
    DefaultParams,
    Event,
    GrammarDefined,
    InputTimersStarted,
    Interpretation,
    InvalidParamValue,
    MethodFailed,
    MethodNotAllowed,
    MethodNotValid,
    MissingParam,
    Opened,
    ParamsSet,
    RecognitionComplete,
    RecognitionInProgress,
    RecogResult,
    StartOfInput,
    Stopped,
    Transcript,
)

SERVER = os.getenv("UHLIVE_API_URL", "wss://api.uh.live")


def build_connection_request(token) -> Tuple[str, dict]:
    """
    Make an authenticated URL and header to connect to the H2B Service.
    """
    return urljoin(SERVER, "/bots"), {"Authorization": f"bearer {token}"}


__all__ = [
    "ProtocolError",
    "Recognizer",
    "build_connection_request",
    "Event",
    "CompletionCause",
    "Transcript",
    "Interpretation",
    "RecogResult",
    "Opened",
    "ParamsSet",
    "DefaultParams",
    "GrammarDefined",
    "RecognitionInProgress",
    "InputTimersStarted",
    "Stopped",
    "Closed",
    "StartOfInput",
    "RecognitionComplete",
    "MethodNotValid",
    "MethodFailed",
    "InvalidParamValue",
    "MissingParam",
    "MethodNotAllowed",
]



"""
Object oriented abstraction over the Conversation API protocol and workflow.
"""

import json
from array import array
from enum import Enum
from typing import Any, Dict, Union

from .events import Event, Ok, SpeakerLeft

# *** Phoenix channel protocol V2 ***
#
# This is a non-official "specification" translated from Phoenix source code.
#
# Compared to the V1 protocol.
#
# * A new message field has been added: `join_ref`, which is the `ref` string when joining. We use "0".
# * JSON messages: instead of an object that names the fields, V2 sends a list
#   (i.e parameters are passed by position instead of being passed by name) in this order:
#      join_ref, ref, topic, event_name, payload
# * We now have a binary version represented as (using erlang binary patterns):
# <<
#  @push::size(8),
#  join_ref_size::size(8),
#  ref_size::size(8),
#  topic_size::size(8),
#  event_size::size(8),
#  join_ref::binary-size(join_ref_size),
#  ref::binary-size(ref_size),
#  topic::binary-size(topic_size),
#  event::binary-size(event_size),
#  data::binary
# >>
# where @push = 0

B_JOIN_REF = ord("1")
S_JOIN_REF = "1"


class ProtocolError(RuntimeError):
    """Exception raised when a [Conversation][uhlive.stream.conversation.Conversation] method is not available in the current state."""

    pass


class State(Enum):
    """Protocol state."""

    Idle = "Idle State"
    Joined = "Joined State"


class Conversation:
    """To join a conversation on the API, you need a `Conversation` object.

    You can only have one `Conversation` per connection (socket) otherwise you risk
    unexpected behavior (and exceptions!).
    """

    def __init__(self, identifier: str, conversation_id: str, speaker: str) -> None:
        """Create a `Conversation`.

        Args:
            identifier: is the identifier you got when you subscribed to the service;
            conversation_id: is the conversation you wish to join,
            speaker: is your alias in the conversation, to identify you and your events
        """
        self._state: State = State.Idle
        self.identifier = identifier
        self.topic = f"conversation:{self.identifier}@{conversation_id}"
        self.topic_bin = self.topic.encode("utf-8")
        self.topic_len = len(self.topic_bin)
        self.speaker = speaker
        self._request_id = int(S_JOIN_REF) - 1

    def join(
        self,
        model: str = "fr",
        country: str = "fr",
        readonly: bool = False,
        interim_results: bool = True,
        rescoring: bool = True,
        origin: int = 0,
        audio_codec: str = "linear",
    ) -> str:
        """Join the conversation.

        Args:
            readonly: if you are not going to stream audio, set it to `True`.
            model: (if `readonly` is `False`) the ASR language model to be use to recognize
                     the audio you will stream.
            country: the iso 2 letter country code of the place where the speaker is.
            interim_results: (`readonly` = `False` only) should the ASR trigger interim result events?
            rescoring: (`readonly` = `False` only) should the ASR refine the final segment
                         with a bigger Language Model?
                         May give slightly degraded results for very short segments.
            origin: The UNIX time, in milliseconds, to which the event timeline origin is set.
            audio_codec: the speech audio codec of the audio data:

                - `"linear"`: (default) linear 16 bit SLE raw PCM audio at 8khz;
                - `"g711a"`: G711 a-law audio at 8khz;
                - `"g711u"`: G711 μ-law audio at 8khz.

        Returns:
            The text websocket message to send to the server.

        Raises:
            ProtocolError: if still in a previously joined conversation.
        """
        if self._state != State.Idle:
            raise ProtocolError("Can't join twice!")
        if not readonly and not model:
            raise ProtocolError("If readonly is False, you must specify a model!")
        return self.command(
            "phx_join",
            {
                "readonly": readonly,
                "speaker": self.speaker,
                "model": model,
                "country": country,
                "interim_results": interim_results,
                "rescoring": rescoring,
                "origin": origin,
                "audio_codec": audio_codec,
            },
        )

    def leave(self) -> str:
        """Leave the current conversation.

        It's a good idea to leave a conversation and continue to consume messages
        until you receive a [`SpeakerLeft`][uhlive.stream.conversation.SpeakerLeft] event for your speaker, before you
        close the connection. Otherwise, you may miss parts of the transcription.

        Returns:
            The text websocket message to send to the server.

        Raises:
            ProtocolError: if not currently in a converstation.
        """
        if self._state != State.Joined:
            raise ProtocolError("No conversation to leave!")
        return self.command("phx_leave", {})

    def send_audio_chunk(self, chunk: bytes) -> bytes:
        """Build an audio chunk for streaming.

        Returns:
            The binary websocket message to send to the server.
        Raises:
            ProtocolError: if not currently in a converstation.
        """
        if self._state != State.Joined:
            raise ProtocolError("Not in a conversation!")
        ref = self.request_id.encode("ascii")
        message = array("B", [0, 1, len(ref), self.topic_len, 11, B_JOIN_REF])
        message.extend(ref)
        message.extend(self.topic_bin)
        message.extend(b"audio_chunk")
        message.extend(chunk)
        return message.tobytes()

    @property
    def request_id(self) -> str:
        self._request_id += 1
        return str(self._request_id)

    def receive(self, data: Union[str, bytes]) -> Event:
        """Decode received websocket message.

        The server only sends text messages.

        Returns:
            The appropriate [Event][uhlive.stream.conversation.Event] subclass instance.
        """
        event = Event.from_message(json.loads(data))
        assert (
            event.topic == self.topic
        ), "Topic mismatch! Are you trying to mix several conversations on the same socket? This is not supported."
        if isinstance(event, Ok) and event.ref == event.join_ref:
            self._state = State.Joined
        elif isinstance(event, SpeakerLeft) and event.speaker == self.speaker:
            self._state = State.Idle
        return event

    @property
    def left(self):
        """Did the server confirm we left the conversation?"""
        return self._state == State.Idle

    def command(self, name: str, payload: Dict[str, Any] = {}) -> str:
        message = [
            S_JOIN_REF,
            self.request_id,
            self.topic,
            name,
            payload,
        ]
        return json.dumps(
            message, ensure_ascii=False, indent=None, separators=(",", ":")
        )



"""H2B Events.

Data model for events returned by the H2B server.
See also https://docs.allo-media.net/stream-h2b/protocols/websocket/#websocket-for-voicebots.
"""

import json
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional


class CompletionCause(Enum):
    """The set of possible completion causes.

    See [all possible values](https://docs.allo-media.net/stream-h2b/protocols/websocket/#asynchronous-recognition-events).
    """

    GramDefinitionFailure = "GramDefinitionFailure"
    GramLoadFailure = "GramLoadFailure"
    Error = "Error"
    LanguageUnsupported = "LanguageUnsupported"
    NoInputTimeout = "NoInputTimeout"
    HotwordMaxtime = "HotwordMaxtime"
    NoMatch = "NoMatch"
    NoMatchMaxtime = "NoMatchMaxtime"
    TooMuchSpeechTimeout = "TooMuchSpeechTimeout"
    PartialMatchMaxtime = "PartialMatchMaxtime"
    Success = "Success"
    PartialMatch = "PartialMatch"


class Transcript:
    """The Transcript part of a recognition result"""

    def __init__(self, data: Dict[str, Any]) -> None:
        self._transcript: str = data["transcript"]
        self._confidence: float = float(data["confidence"])
        self._start = datetime.utcfromtimestamp(data["start"] / 1000.0)
        self._end = datetime.utcfromtimestamp(data["end"] / 1000.0)

    @property
    def transcript(self) -> str:
        """The raw ASR output."""
        return self._transcript

    @property
    def confidence(self) -> float:
        """The ASR transcription confidence."""
        return self._confidence

    @property
    def start(self) -> datetime:
        """Start of speech."""
        return self._start

    @property
    def end(self) -> datetime:
        """End of speech."""
        return self._end

    def __str__(self) -> str:
        return f'"{self._transcript}" ({self._confidence})'


class Interpretation:
    """The interpretation part of a recognition result"""

    def __init__(self, data: Dict[str, Any]) -> None:
        self._type: str = data["type"]
        self._value: Dict[str, Any] = data["value"]
        self._confidence: float = float(data["confidence"])

    @property
    def confidence(self) -> float:
        """The confidence of the interpretation."""
        return self._confidence

    @property
    def type(self) -> str:
        """The type of the Interpretation is given by the builtin grammar URI"""
        return self._type

    @property
    def value(self) -> Dict[str, Any]:
        """The structured interpreted value.

        The type/schema of the value is given by the `self.type` property.

        See the [Grammar reference documentaiton](https://docs.allo-media.net/stream-h2b/grammars).
        """
        return self._value

    def __str__(self) -> str:
        return f"[{self._type}] {self._value} ({self._confidence})"


class RecogResult:
    """When a recognition completes, this describe the result"""

    def __init__(self, data: dict) -> None:
        self._asr: Optional[Transcript] = (
            None if not data["asr"] else Transcript(data["asr"])
        )
        self._nlu: Optional[Interpretation] = (
            None if not data["nlu"] else Interpretation(data["nlu"])
        )
        self._grammar_uri: str = data["grammar_uri"]
        if "alternatives" in data:
            self._alternatives = [RecogResult(alt) for alt in data["alternatives"]]

    @property
    def asr(self) -> Optional[Transcript]:
        """The ASR part of the result ([transcription][uhlive.stream.recognition.Transcript] result)"""
        return self._asr

    @property
    def nlu(self) -> Optional[Interpretation]:
        """The NLU part of the result ([interpretation][uhlive.stream.recognition.Interpretation])"""
        return self._nlu

    @property
    def grammar_uri(self) -> str:
        """The grammar that matched, as it was given to the `RECOGNIZE` command"""
        return self._grammar_uri

    @property
    def alternatives(self) -> Optional[List["RecogResult"]]:
        """if N-bests were requested, the additional results besides the best one are there."""
        return getattr(self, "_alternatives", None)

    def __str__(self) -> str:
        best = f"Transcript: {self._asr}\n NLU: {self._nlu}"
        if hasattr(self, "_alternatives"):
            alt_str = "\n   ".join(str(alt) for alt in self._alternatives)
            return f"{best}\nAlternatives:\n   {alt_str}"
        else:
            return best


class Event:
    """Base class of all the events"""

    def __init__(self, data: Dict[str, Any]) -> None:
        self._request_id: int = int(data["request_id"])
        self._channel_id: str = data["channel_id"]
        self._headers: Dict[str, Any] = data["headers"]
        self._completion_cause: Optional[CompletionCause] = (
            None
            if not data["completion_cause"]
            else CompletionCause(data["completion_cause"])
        )
        self._completion_reason: Optional[str] = data["completion_reason"] or None
        self._body: Optional[RecogResult] = (
            None if not data["body"] else RecogResult(data["body"])
        )

    @property
    def request_id(self) -> int:
        """The request ID that event responds to."""
        return self._request_id

    @property
    def channel_id(self) -> str:
        """The channel ID."""
        return self._channel_id

    @property
    def headers(self) -> Dict[str, Any]:
        """The response headers.

        See also the [header description](https://docs.allo-media.net/stream-h2b/output/#headers-%26-statuses).
        """
        return self._headers

    @property
    def completion_cause(self) -> Optional[CompletionCause]:
        """The response [`CompletionCause`][uhlive.stream.recognition.CompletionCause]."""
        return self._completion_cause

    @property
    def completion_reason(self) -> Optional[str]:
        """The completion message."""
        return self._completion_reason

    @property
    def body(self) -> Optional[RecogResult]:
        """The content of the Event is a [`RecogResult`][uhlive.stream.recognition.RecogResult] if it is a `RecognitionComplete` event."""
        return self._body

    def __str__(self) -> str:
        return f"<Event {self.__class__.__name__}: {self._completion_cause} – {self._completion_reason} – {self._headers} – {self._body}"


class Opened(Event):
    """Session opened on the server"""

    pass


class ParamsSet(Event):
    """The default parameters were set."""

    pass


class DefaultParams(Event):
    """All the parameters and their values are in the `headers` property"""

    pass


class GrammarDefined(Event):
    """The `DefineGrammar` command has been processed."""

    pass


class RecognitionInProgress(Event):
    """The ASR recognition is started."""

    pass


class InputTimersStarted(Event):
    """The Input Timers are started."""

    pass


class Stopped(Event):
    """The ASR recognition has been stopped on the client request."""

    pass


class Closed(Event):
    """The session is closed."""

    pass


class StartOfInput(Event):
    """In normal recognition mode, this event is emitted when speech is detected."""

    pass


class RecognitionComplete(Event):
    """The ASR recognition is complete."""

    pass


class MethodNotValid(Event):
    """The server received an invalid command."""

    pass


class MethodFailed(Event):
    """The server was unable to complete the command."""

    pass


class InvalidParamValue(Event):
    """The server received a request to set an invalid value for a parameter."""

    pass


class MissingParam(Event):
    """The command is missings some mandatory parameter."""

    pass


class MethodNotAllowed(Event):
    """The command is not allowed in this state."""

    pass


EVENT_MAP = {
    "OPENED": Opened,
    "PARAMS-SET": ParamsSet,
    "DEFAULT-PARAMS": DefaultParams,
    "GRAMMAR-DEFINED": GrammarDefined,
    "RECOGNITION-IN-PROGRESS": RecognitionInProgress,
    "INPUT-TIMERS-STARTED": InputTimersStarted,
    "STOPPED": Stopped,
    "CLOSED": Closed,
    "START-OF-INPUT": StartOfInput,
    "RECOGNITION-COMPLETE": RecognitionComplete,
    "METHOD-NOT-VALID": MethodNotValid,
    "METHOD-FAILED": MethodFailed,
    "INVALID-PARAM-VALUE": InvalidParamValue,
    "MISSING-PARAM": MissingParam,
    "METHOD-NOT-ALLOWED": MethodNotAllowed,
}


def deserialize(data: str) -> Event:
    jd = json.loads(data)
    kind = jd["event"]
    if kind in EVENT_MAP:
        return EVENT_MAP[kind](jd)
    raise ValueError(f"Unknown event '{kind}'")



"""
Display helpers.
"""

from datetime import datetime

local_tz = datetime.now().astimezone().tzinfo


def human_datetime(timestamp):
    """Human readable representation of unix timestamp date."""
    utc_dt = datetime.fromtimestamp(timestamp / 1000.0, local_tz)

    return utc_dt.isoformat(sep=" ")



from unittest import TestCase

from uhlive.stream.conversation.events import (
    EntityFound,
    Event,
    SegmentDecoded,
    SpeakerJoined,
    SpeakerLeft,
    WordsDecoded,
)

from .conversation_events import (
    entity_location_city_found,
    entity_number_found,
    segment_decoded,
    speaker_joined,
    speaker_left,
    words_decoded,
)


class TestConversationEventReceived(TestCase):
    def test_creation(self):
        message = words_decoded
        Event.from_message(message)

    def test_speaker(self):
        message = words_decoded
        event = Event.from_message(message)

        self.assertEqual(event.speaker, "robin")

        message = speaker_left
        event = Event.from_message(message)

        self.assertEqual(event.speaker, "robin")

    def test_transcript(self):
        message = words_decoded
        event = Event.from_message(message)

        self.assertEqual(event.transcript, "allez-y s' il vous plaît de participer")

    def test_timestamps(self):
        event = Event.from_message(words_decoded)
        self.assertEqual(event.start, 2310)
        self.assertEqual(event.end, 5670)
        self.assertEqual(event.length, 3360)
        event = Event.from_message(entity_number_found)
        self.assertEqual(event.start, 2730)
        event = Event.from_message(speaker_left)
        self.assertEqual(event.timestamp, 1613129073523)

    def test_lang(self):
        event = Event.from_message(words_decoded)
        self.assertEqual(event.lang, "fr")
        event = Event.from_message(entity_number_found)
        self.assertEqual(event.lang, "fr")

    def test_type(self):
        message = words_decoded
        event = Event.from_message(message)
        self.assertIsInstance(event, WordsDecoded)

        event = Event.from_message(segment_decoded)
        self.assertIsInstance(event, SegmentDecoded)

        message = speaker_left
        event = Event.from_message(message)
        self.assertIsInstance(event, SpeakerLeft)

        event = Event.from_message(entity_number_found)
        self.assertIsInstance(event, EntityFound)
        self.assertEqual(event.entity_name, "number")

    def test_annotation(self):
        annotation = Event.from_message(entity_number_found)
        self.assertEqual(annotation.canonical, "22")
        self.assertEqual(annotation.value, 22.0)
        self.assertEqual(annotation.original, "vingt-deux")

    def test_words(self):
        message = words_decoded
        event = Event.from_message(message)

        self.assertEqual(
            event.words,
            [
                {
                    "confidence": 0.94202,
                    "start": 960,
                    "end": 1320,
                    "length": 360,
                    "word": "allez-y",
                },
                {
                    "confidence": 0.954795,
                    "start": 1320,
                    "end": 1350,
                    "length": 30,
                    "word": "s'",
                },
                {
                    "confidence": 0.999226,
                    "start": 1350,
                    "end": 1440,
                    "length": 90,
                    "word": "il",
                },
                {
                    "confidence": 0.999994,
                    "start": 1440,
                    "end": 1560,
                    "length": 120,
                    "word": "vous",
                },
                {
                    "confidence": 0.99975,
                    "start": 1560,
                    "end": 1740,
                    "length": 180,
                    "word": "plaît",
                },
                {
                    "confidence": 0.663816,
                    "start": 1740,
                    "end": 1770,
                    "length": 30,
                    "word": "de",
                },
                {
                    "confidence": 0.852544,
                    "start": 1770,
                    "end": 2700,
                    "length": 930,
                    "word": "participer",
                },
            ],
        )

        first = event.words[0]
        for attr in ["start", "end", "length", "word", "confidence"]:
            self.assertEqual(getattr(first, attr), first[attr])

    def test_speaker_joined(self):
        event = Event.from_message(speaker_joined)
        self.assertIsInstance(event, SpeakerJoined)
        self.assertEqual(event.speaker, "robin")
        self.assertTrue(event.rescoring)
        self.assertTrue(event.interim_results)
        self.assertEqual(event.timestamp, 1613129063523)

    def test_instantiate_ner_event(self):
        event = Event.from_message(entity_location_city_found)
        self.assertIsInstance(event, EntityFound)
        self.assertEqual(event.entity_name, "location_city")
        self.assertEqual(event.original, "lyon")
        self.assertEqual(event.confidence, 0.9)



"""
Object oriented abstraction over the H2B API protocol and workflow.
"""

import json
from enum import Enum
from typing import Any, Dict, Union

from .events import (
    Closed,
    Event,
    Opened,
    RecognitionComplete,
    RecognitionInProgress,
    Stopped,
    deserialize,
)


def serialize(cmd: Dict[str, Any]) -> str:
    return json.dumps(cmd, ensure_ascii=False, indent=None, separators=(",", ":"))


class ProtocolError(RuntimeError):
    """Exception raised when a [Recognizer][uhlive.stream.recognition.Recognizer] method is not available in the current state."""

    pass


class State(Enum):
    """Protocol state."""

    NoSession = "Out of Session State"
    IdleSession = "Idle Session State"
    Recognition = "On-going Recognition State"


class Recognizer:
    """The connection state machine.

    Use this class to decode received frames as `Event`s or to
    make command frames by calling the appropriate methods.
    If you call a method that is not appropriate in the current protocol
    state, a `ProtocolError` is raised.
    """

    def __init__(self) -> None:
        self._state: State = State.NoSession
        self._request_id = 0
        self._channel_id = ""

    # Workflow methods

    def open(
        self,
        custom_id: str = "",
        channel_id: str = "",
        session_id: str = "",
        audio_codec: str = "linear",
    ) -> str:
        """Open a new H2B session.

        Args:
            custom_id: is any reference of yours that you want to appear in the logs
                       and invoice reports; for example your client id.
            channel_id: when provided, it'll be used as a prefix for
                        the actual channel ID generated by the server.
            session_id: you can set an alternate channel identifier to appear along side the channel_id in
                        the logs and in the Dev Console, to make it easier to reconcile your client logs with
                        the server logs.
            audio_codec: the speech audio codec of the audio data:

                - `"linear"`: (default) linear 16 bit SLE raw PCM audio at 8khz;
                - `"g711a"`: G711 a-law audio at 8khz;
                - `"g711u"`: G711 μ-law audio at 8khz.

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if a session is already open.
        """
        if self._state != State.NoSession:
            raise ProtocolError("Session already opened!")
        return serialize(
            {
                "command": "OPEN",
                "request_id": self.request_id,
                "channel_id": channel_id,
                "headers": {
                    "custom_id": custom_id,
                    "audio_codec": audio_codec,
                    "session_id": session_id,
                },
                "body": "",
            }
        )

    def send_audio_chunk(self, chunk: bytes) -> bytes:
        """Build an audio chunk frame for streaming.

        Returns:
            A websocket binary message to send to the server.

        Raises:
            ProtocolError: if no session opened.
        """
        if self._state == State.NoSession:
            raise ProtocolError("You must open a session first!")
        return chunk

    def set_params(self, **params: Any) -> str:
        """Set default ASR parameters for the session.

        See [the parameter list](https://docs.allo-media.net/stream-h2b/protocols/websocket/#set-session-defaults)
        and [the parameter visual explanations](https://docs.allo-media.net/stream-h2b/input/#resource-headers)
        for an explanation of the different parameters available.

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if no session opened.
        """
        if self._state != State.IdleSession:
            raise ProtocolError(f"Method not available in this state ({self._state})!")
        return self.command("SET-PARAMS", params)

    def get_params(self) -> str:
        """Retrieve the default values for the ASR parameters.

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if no session opened.
        """
        if self._state != State.IdleSession:
            raise ProtocolError(f"Method not available in this state ({self._state})!")
        return self.command("GET-PARAMS")

    def define_grammar(self, builtin: str, alias: str) -> str:
        """Define a grammar alias for a parameterized builtin.

        Args:
            builtin: the builtin URI to alias, including the query string, but without the "builtin:" prefix
            alias: the alias, without the "session:" prefix.

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if no session opened.
        """
        if self._state != State.IdleSession:
            raise ProtocolError(f"Method not available in this state ({self._state})!")
        return self.command(
            "DEFINE-GRAMMAR",
            {"content_id": alias, "content_type": "text/uri-list"},
            f"builtin:{builtin}",
        )

    def recognize(
        self,
        *grammars: str,
        start_timers: bool = True,
        recognition_mode: str = "normal",
        **params: Any,
    ) -> str:
        """Start a recognition process.

        This method takes grammar URIs as positional arguments, including the `builtin:` or
        `session:` prefixes to make the difference between builtin grammars and custom aliases.

        Keyword Args:
          start_timers: default True
          recognition_mode: default is "normal"
          **params: any other [ASR parameter](https://docs.allo-media.net/stream-h2b/protocols/websocket/#start-recognition) (no client side defaults).

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if no session opened.
        """
        if self._state != State.IdleSession:
            raise ProtocolError(f"Method not available in this state ({self._state})!")
        return self.command(
            "RECOGNIZE",
            headers={
                "recognition_mode": recognition_mode,
                "content_type": "text/uri-list",
                "start_input_timers": start_timers,
                **params,
            },
            body="\n".join(grammars),
        )

    def close(self) -> str:
        """Close the current session.

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if no session opened.
        """
        if self._state != State.IdleSession:
            raise ProtocolError(f"Method not available in this state ({self._state})!")
        return self.command("CLOSE")

    def start_input_timers(self) -> str:
        """If the input timers were not started by the RECOGNIZE command,
        starts them now.

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if no on-going recognition process
        """
        if self._state != State.Recognition:
            raise ProtocolError("Command is only valid during recognition!")
        return self.command("START-INPUT-TIMERS")

    def stop(self) -> str:
        """Stop ongoing recognition process

        Returns:
            A websocket text message to send to the server.

        Raises:
            ProtocolError: if no on-going recognition process
        """
        if self._state != State.Recognition:
            raise ProtocolError("Command is only valid during recognition!")
        return self.command("STOP")

    ##

    @property
    def request_id(self) -> int:
        self._request_id += 1
        return self._request_id

    @property
    def channel_id(self) -> str:
        """The current session ID."""
        return self._channel_id

    def receive(self, data: Union[str, bytes]) -> Event:
        """Decode received text frame.

        The server always replies with text frames.

        Returns:
            The appropriate `Event` subclass.
        """
        assert type(data) is str  # to please mypy
        event = deserialize(data)
        if isinstance(event, RecognitionInProgress):
            self._state = State.Recognition
        elif isinstance(event, (RecognitionComplete, Stopped)):
            self._state = State.IdleSession
        elif isinstance(event, Opened):
            self._channel_id = event.channel_id
            self._state = State.IdleSession
        elif isinstance(event, Closed):
            self._state = State.NoSession
        return event

    def command(self, name: str, headers: Dict[str, Any] = {}, body: str = "") -> str:
        return serialize(
            {
                "command": name,
                "request_id": self.request_id,
                "channel_id": self.channel_id,
                "headers": headers,
                "body": body,
            }
        )



from unittest import TestCase

from uhlive.stream.recognition import Opened, ProtocolError, Recognizer

from .recog_events import recognition_complete, recognition_in_progress, session_opened


class TestConnection(TestCase):
    def test_open(self):
        client = Recognizer()
        frame = client.open("mycustomer", "mychan")
        self.assertEqual(
            frame,
            r'{"command":"OPEN","request_id":1,"channel_id":"mychan","headers":{"custom_id":"mycustomer","audio_codec":"linear","session_id":""},"body":""}',
        )
        with self.assertRaises(ProtocolError):
            client.get_params()
        with self.assertRaises(ProtocolError):
            client.send_audio_chunk(b"trntrntrntnrtsrn")

    def test_opened(self):
        client = Recognizer()
        event = client.receive(session_opened)
        self.assertIsInstance(event, Opened)
        frame = client.get_params()
        self.assertEqual(
            frame,
            r'{"command":"GET-PARAMS","request_id":1,"channel_id":"testuie46e4ui6","headers":{},"body":""}',
        )
        # Can't open twice
        with self.assertRaises(ProtocolError):
            client.open("test", "test")

        # we can stream
        client.send_audio_chunk(b"trntrntrntnrtsrn")

    def test_recognize(self):
        client = Recognizer()
        client.receive(session_opened)
        frame = client.define_grammar(
            "speech/spelling/mixed?regex=[a-z]{2}[0-9]{9}[a-z]{2}", "parcel_num"
        )
        self.assertEqual(
            frame,
            r'{"command":"DEFINE-GRAMMAR","request_id":1,"channel_id":"testuie46e4ui6","headers":{"content_id":"parcel_num","content_type":"text/uri-list"},"body":"builtin:speech/spelling/mixed?regex=[a-z]{2}[0-9]{9}[a-z]{2}"}',
        )
        frame = client.recognize("session:parcel_num", no_input_timeout=5000)
        self.assertEqual(
            frame,
            r'{"command":"RECOGNIZE","request_id":2,"channel_id":"testuie46e4ui6","headers":{"recognition_mode":"normal","content_type":"text/uri-list","start_input_timers":true,"no_input_timeout":5000},"body":"session:parcel_num"}',
        )

    def test_recognition_in_progress(self):
        client = Recognizer()
        client.receive(session_opened)
        client.receive(recognition_in_progress)
        # Can't have multiple recognition processes
        with self.assertRaises(ProtocolError):
            client.recognize("builtin:speech/transcribe")

        # Only stop and start-input-timers are valid
        with self.assertRaises(ProtocolError):
            client.get_params()

        self.assertEqual(
            client.start_input_timers(),
            r'{"command":"START-INPUT-TIMERS","request_id":1,"channel_id":"testuie46e4ui6","headers":{},"body":""}',
        )

        self.assertEqual(
            client.stop(),
            r'{"command":"STOP","request_id":2,"channel_id":"testuie46e4ui6","headers":{},"body":""}',
        )

        # we can stream too
        client.send_audio_chunk(b"trntrntrntnrtsrn")

    def test_set_params(self):
        client = Recognizer()
        client.receive(session_opened)
        self.assertEqual(
            client.set_params(speech_language="fr", confidence_threshold=0.7),
            r'{"command":"SET-PARAMS","request_id":1,"channel_id":"testuie46e4ui6","headers":{"speech_language":"fr","confidence_threshold":0.7},"body":""}',
        )

    def test_recognition_complete(self):
        client = Recognizer()
        client.receive(session_opened)
        client.receive(recognition_in_progress)
        with self.assertRaises(ProtocolError):
            client.recognize("builtin:speech/transcribe")
        client.receive(recognition_complete)
        # Now we can start another recognition process
        client.recognize("builtin:speech/transcribe")



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="uhlive",
    version="1.5.0",
    url="https://github.com/uhlive/python-sdk",
    author="Allo-Media",
    author_email="support@allo-media.fr",
    description="Python bindings for the Uh!ive APIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    extras_require={
        "examples": ["websocket-client", "requests", "aiohttp", "sounddevice", "toml"]
    },
    python_requires=">=3.7",
)



"""
Authentication helpers
"""

import os
from typing import Dict, Tuple

SERVER = os.getenv("UHLIVE_AUTH_SERVER", "id.uh.live")
REALM = os.getenv("UHLIVE_AUTH_REALM", "uhlive")


def build_authentication_request(
    client_id: str, client_secret: str, user_id: str = "", user_pwd: str = ""
) -> Tuple[str, Dict[str, str]]:
    """Build a couple (url, param dict) to be use in an HTTP POST request to get the API access_token.

    Synchronous example using ``requests``:

    ```python
    auth_url, auth_params = build_authentication_request(uhlive_client, uhlive_secret)
    login = requests.post(auth_url, data=auth_params)
    login.raise_for_status()
    uhlive_token = login.json()["access_token"]
    ```
    """

    url = f"https://{SERVER}/realms/{REALM}/protocol/openid-connect/token"
    data = {
        "client_id": client_id,
        "grant_type": "client_credentials" if not user_id else "password",
        "client_secret": client_secret,
    }
    if user_id:
        data["username"] = user_id
        data["password"] = user_pwd
    return url, data
